// Rendezvous: A Hopfield-Tank neural network implementation as a scheduler application
// Written by David J. Stein, Esq., 10/18/2004.  See "License.html" for terms and conditions of use.
//
// This module contains all of the code for the neural network embedded in Rendezvous.

#region Using declarations
using System;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Schema;
using System.Collections;
#endregion

namespace Hopfield_Tank_Neural_Network__Rendezvous_Namespace {

	#region Resource classes: Person, Meeting, Room, and TimeSlot
	class Person {
		internal string name;
		internal Person(string name) {
			this.name = name;
		}
	}

	class Meeting {
		internal string name;
		internal Meeting(string name) {
			this.name = name;
		}
	}

	class Room {
		internal string name;
		internal int capacity;
		internal Room(string name, int capacity) {
			this.name = name;
			this.capacity = capacity;
		}
	}

	class TimeSlot {
		internal string name;
		internal TimeSlot(string name) {
			this.name = name;
		}
	}
	#endregion

	public class HopfieldTankNetwork	{

		#region Variables and initialization code

		// control variables
		internal string name = "Sample Network";
		internal ArrayList people;     // list of Person objects
		internal ArrayList meetings;   // list of Meeting objects
		internal ArrayList rooms;      // list of Room objects
		internal ArrayList timeslots;  // list of time slot names
		internal bool[,] rsvp;         // person, meeting

		// network variables
		internal double[,,] V;        // time slot, room, meeting
		internal double[,,] VPrevious;
		internal double[,,] u;
		internal double[,,] uInitial;
		internal double u0 = 0.5;
		internal double du0 = 0.01;
		internal double dt = 0.0001;
		internal double E = 1000000;
		internal double EPrevious = 1000001;
		internal Random random;
		internal string XmlErrors = "";

		internal HopfieldTankNetwork(int seed) {
   random = new Random(seed);
			people = new ArrayList();
			meetings = new ArrayList();
			timeslots = new ArrayList();
			rooms = new ArrayList();
			rsvp = new bool[0,0];
			Initialize();
		}

		#endregion

		#region Core functions

		internal void Initialize() {
			// This function sets V and u values, and stores the uInitial values for later use in Reset.
			V = new double[timeslots.Count, rooms.Count, meetings.Count];
   VPrevious = new double[timeslots.Count, rooms.Count, meetings.Count];
			u = new double[timeslots.Count, rooms.Count, meetings.Count];
			uInitial = new double[timeslots.Count, rooms.Count, meetings.Count];
			for (int t = 0; t < timeslots.Count; t++) {
				for (int r = 0; r < rooms.Count; r++) {
					for (int m = 0; m < meetings.Count; m++) {
						V[t, r, m] = VPrevious[t, r, m] = 0.0;
      u[t, r, m] = uInitial[t, r, m] = u0 + (random.Next(100) / 100.0) * (2 * du0) - du0;
					}
				}
			}
		}

		internal void Reset() {
			// This function resets all of the u (and V) values to their initial state for replay.
			for (int t = 0; t < timeslots.Count; t++) {
				for (int r = 0; r < rooms.Count; r++) {
					for (int m = 0; m < meetings.Count; m++) {
						V[t, r, m] = VPrevious[t, r, m] = 0.0;
						u[t, r, m] = uInitial[t, r, m];
					}
				}
			}
		}

		internal void AddPerson(string name) {
			// This function adds a person to the network and resizes teh RSVP array.
			people.Add(new Person(name));
			// create new rsvp array, copy previous entries, and insert 0's for new person
   bool[,] newrsvp = new bool[meetings.Count, people.Count];
			for (int m = 0; m < meetings.Count; m++) {
				for (int p = 0; p < people.Count - 1; p++) {
					newrsvp[m, p] = rsvp[m, p];
				}
			}
			for (int m = 0; m < meetings.Count; m++)
				newrsvp[m, people.Count - 1] = false;
			rsvp = newrsvp;
			Initialize();
		}

		internal void RemovePerson(int index) {
			// This function removes a person from the network and resizes the RSVP array.
   people.RemoveAt(index);
			// create new rsvp array and copy all entries except those for deleted person
			bool[,] newrsvp = new bool[meetings.Count, people.Count];
			for (int m = 0; m < meetings.Count; m++) {
				for (int p = 0; p < index - 1; p++) {
					newrsvp[m, p] = rsvp[m, p];
				}
				for (int p = index; p < people.Count; p++) {
					newrsvp[m, p] = rsvp[m, p + 1];
				}
			}
			rsvp = newrsvp;
			Initialize();
		}

		internal void AddMeeting(string name) {
			// This function adds a meeting to the network and resizes the RSVP array.
			meetings.Add(new Meeting(name));
			// create new rsvp array, copy previous entries, and insert 0's for new meeting
			bool[,] newrsvp = new bool[meetings.Count, people.Count];
			for (int m = 0; m < meetings.Count - 1; m++) {
				for (int p = 0; p < people.Count; p++) {
					newrsvp[m, p] = rsvp[m, p];
				}
			}
			for (int p = 0; p < people.Count; p++)
				newrsvp[meetings.Count - 1, p] = false;
			rsvp = newrsvp;
			Initialize();
		}

		internal void RemoveMeeting(int index) {
			// This function removes a meeting from the network and resizes the RSVP array.
			meetings.RemoveAt(index);
			// create new rsvp array and copy all entries except those for deleted meeting
			bool[,] newrsvp = new bool[meetings.Count, people.Count];
			for (int p = 0; p < people.Count; p++) {
				for (int m = 0; m < index; m++) {
					newrsvp[m, p] = rsvp[m, p];
				}
				for (int m = index; m < meetings.Count; m++) {
					newrsvp[m, p] = rsvp[m + 1, p];
				}
			}
			rsvp = newrsvp;
			Initialize();
		}

		internal void AddRoom(string name, int capacity) {
			// This function adds a room to the network.
			rooms.Add(new Room(name, capacity));
			Initialize();
		}

		internal void RemoveRoom(int index) {
			// This function removes a room from the network.
			rooms.RemoveAt(index);
			Initialize();
		}

		internal void AddTimeSlot(string name) {
			// This function adds a time slot to the network.
			timeslots.Add(new TimeSlot(name));
			Initialize();
		}

		internal void RemoveTimeSlot(int index) {
			// This function removes a time slot from the network.
			timeslots.RemoveAt(index);
			Initialize();
		}

		internal void Analyze() {
			// This function is the heart of the neural network - it adjusts the V and u values, and computes a new E value.

			// adjust V values

			for (int t = 0; t < timeslots.Count; t++) {
				for (int r = 0; r < rooms.Count; r++) {
					for (int m = 0; m < meetings.Count; m++) {
						VPrevious[t, r, m] = V[t, r, m];
						V[t,r,m] = 0.5 * (1.0 + Math.Tanh(u[t, r, m] / u0));
					}
				}
			}

			// calculate du values

			// find maximum participation in any meeting
			int maxcount = 0;
			for (int m = 0; m < meetings.Count; m++) {
				int count = 0;
				for (int p = 0; p < people.Count; p++) {
					if (rsvp[m, p] == true)
						count++;
				}
				maxcount = Math.Max(maxcount, count);
			}

			for (int m = 0; m < meetings.Count; m++) {
				for (int t = 0; t < timeslots.Count; t++) {
					for (int r = 0; r < rooms.Count; r++) {

						// first term: increase event probability if this room has sufficient capacity for this event
						double ASum = 0.0;
						int count = 0;
						for (int p = 0; p < people.Count; p++) {
							if (rsvp[m, p] == true)
								count++;
						}
						ASum = (((Room) rooms[r]).capacity >= count ? (-0.9 * (double) count / (double) maxcount) : 0);

      // second term: every meeting can only be held once
						double BSum = 0.0;
						for (int t2 = 0; t2 < timeslots.Count; t2++) {
							for (int r2 = 0; r2 < rooms.Count; r2++) {
								if ((t != t2) || (r != r2))
									BSum += V[t2, r2, m];
							}
						}

						// third term: no two meetings can be offered in the same room at the same time
						double CSum = 0.0;
						for (int m2 = 0; m2 < meetings.Count; m2++) {
							if (m != m2)
								CSum += V[t, r, m2];
						}

						// fourth term: every meeting must not be offered at the same time as any other meeting if any one person overlaps
						double DSum = 0.0;
						for (int m2 = 0; m2 < meetings.Count; m2++) {
							if (m != m2) {
								for (int r2 = 0; r2 < rooms.Count; r2++) {
									for (int p = 0; p < people.Count; p++) {
										if ((rsvp[m, p] == true) && (rsvp[m2, p] == true))
											DSum += V[t, r2, m2];
									}
								}
							}
						}

						// calculate du/dt and add into u
						double dudt = -(ASum + BSum + CSum + DSum);
						u[t, r, m] += dudt * dt;
						if (u[t, r, m] < -50.0)
							u[t, r, m] = -50.0;
						else if (u[t, r, m] > 50.0)
							u[t, r, m] = 50.0;
					}
				}
			}

			// calculate E value
			EPrevious = E;
			E = 0.0;

			double EASum = 0.0;
			// first term: every meeting is offered exactly once
			for (int m = 0; m < meetings.Count; m++) {
				for (int t = 0; t < timeslots.Count; t++) {
					for (int r = 0; r < rooms.Count; r++) {
						EASum += V[t, r, m];
					}
				}
			}
			E += 0.5 * Math.Pow(EASum - meetings.Count, 2);

			// second term: no two meetings can be offered in the same room at the same time
			double EBSum = 0.0;
				for (int t = 0; t < timeslots.Count; t++) {
					for (int r = 0; r < rooms.Count; r++) {
						for (int m = 0; m < meetings.Count; m++) {
							for (int m2 = 0; m2 < meetings.Count; m2++) {
								if (m != m2)
									EBSum += V[t, r, m] * V[t, r, m2];
						}
					}
				}
			}
			E += 0.5 * EBSum;

			// third term: check overlap
			double ECSum = 0.0;
				for (int t = 0; t < timeslots.Count; t++) {
					for (int r = 0; r < rooms.Count; r++) {
						for (int m = 0; m < meetings.Count; m++) {
							for (int m2 = 0; m2 < meetings.Count; m2++) {
							if (m != m2) {
								for (int r2 = 0; r2 < rooms.Count; r2++) {
									for (int p = 0; p < people.Count; p++) {
										if ((rsvp[m, p] == true) && (rsvp[m2, p] == true))
											ECSum += V[t, r, m] * V[t, r2, m2];
									}
								}
							}
						}
					}
				}
			}
			E += 0.5 * ECSum;
		}

		internal static HopfieldTankNetwork Load(string filename) {
			// This function loads a network. (Note that it's static, belonging to the class, and simply returns the loaded network.)
			// This function utilizes the Network.xsd file stored (hopefully) in the Schema Files folder. It produces error messages if the
			// selected network file does not fit the network schema.
			HopfieldTankNetwork network = new HopfieldTankNetwork((int) System.DateTime.Now.Ticks);
			try {
				XmlTextReader xtr = new XmlTextReader(filename);
				XmlValidatingReader xvr = new XmlValidatingReader(xtr);
				xvr.ValidationType = ValidationType.Schema;
				xvr.ValidationEventHandler += new ValidationEventHandler(network.ValidationErrors);
				XmlSchemaCollection xsc = new XmlSchemaCollection();
				xsc.Add("", System.Reflection.Assembly.GetExecutingAssembly().Location.Substring(0, System.Reflection.Assembly.GetExecutingAssembly().Location.Length - 45) + @"Schema Files\Network.xsd");
				xvr.Schemas.Add(xsc);
				XmlDocument xd = new XmlDocument();
				network.XmlErrors = "";
				xd.Load(xvr);
				if (network.XmlErrors.Length > 0)
					MessageBox.Show("Errors loading network document. This does not appear to be a valid network file.", "Errors Loading Network Document", MessageBoxButtons.OK, MessageBoxIcon.Error);
				else {
					foreach (XmlNode x1 in xd.ChildNodes) {
						if (x1.Name == "RendezvousHopfieldTankNetwork") {
							foreach (XmlNode x2 in x1.ChildNodes) {
								if (x2.Name == "Parameters") {
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "NetworkName")
											network.name = x3.InnerText;
										else if (x3.Name == "u0")
											network.u0 = double.Parse(x3.InnerText);
										else if (x3.Name == "du0")
											network.du0 = double.Parse(x3.InnerText);
										else if (x3.Name == "dt")
											network.dt = double.Parse(x3.InnerText);
									}
								}
								else if (x2.Name == "People") {
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "Person") {
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "Name") {
													network.people.Add(new Person(x4.InnerText));
												}
											}
										}
									}
								}
								else if (x2.Name == "Meetings") {
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "Meeting") {
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "Name") {
													network.meetings.Add(new Meeting(x4.InnerText));
												}
											}
										}
									}
								}
								else if (x2.Name == "Rooms") {
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "Room") {
											string name = "";
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "Name")
													name = x4.InnerText;
												else if (x4.Name == "Capacity")
													network.rooms.Add(new Room(name, Int32.Parse(x4.InnerText)));
											}
										}
									}
								}
								else if (x2.Name == "TimeSlots") {
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "TimeSlot") {
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "Name") {
													network.timeslots.Add(new TimeSlot(x4.InnerText));
												}
											}
										}
									}
								}
								else if (x2.Name == "RSVP") {
									network.rsvp = new bool[network.meetings.Count, network.people.Count];
									network.Initialize();
									XmlNode x3 = x2.ChildNodes[0];
									for (int m = 0; m < network.meetings.Count; m++) {
										for (int p = 0; p < network.people.Count; p++) {
											network.rsvp[m, p] = false;
											if (x3.Name == "RSVPElement") {
												network.rsvp[m, p] = bool.Parse(x3.InnerText);
												x3 = x3.NextSibling;
											}
										}
									}
								}
								else if (x2.Name == "Network") {
									int t = 0;
									int r = 0;
									int m = 0;
									foreach (XmlNode x3 in x2.ChildNodes) {
										if (x3.Name == "NetworkElement") {
											foreach (XmlNode x4 in x3.ChildNodes) {
												if (x4.Name == "VCurrent")
													network.V[t, r, m] = Double.Parse(x4.InnerText);
												else if (x4.Name == "VPrevious")
													network.VPrevious[t, r, m] = Double.Parse(x4.InnerText);
												else if (x4.Name == "uCurrent")
													network.u[t, r, m] = Double.Parse(x4.InnerText);
												else if (x4.Name == "uInitial") {
													network.uInitial[t, r, m] = Double.Parse(x4.InnerText);
												}
											}
											if ((++r) == network.rooms.Count) {
												r = 0;
												if ((++m) == network.meetings.Count) {
													m = 0;
													if ((++t) == network.meetings.Count)
														t = 0;
												}
											}
										}
									}
								}
							}
						}
					}
					xtr.Close();
				}
			} catch (Exception ex) {
				System.Windows.Forms.MessageBox.Show(ex.Message);
				network.XmlErrors += ex.Message;
			}
			if (network.XmlErrors.Length == 0)
				return network;
			else
				return null;
		}
		public void ValidationErrors(object sender, ValidationEventArgs e) {
			XmlErrors += e.Message + "\r\n";
		}

		internal void Save(string filename) {
			// This function saves the neural network to a file, using good XML practices.
			XmlTextWriter save = new XmlTextWriter(filename, System.Text.Encoding.UTF8);
			save.Formatting = Formatting.Indented;
			save.IndentChar = ' ';
			save.WriteStartDocument();
			save.WriteStartElement("RendezvousHopfieldTankNetwork");
			save.Indentation++;
			// write parameters
			save.WriteStartElement("Parameters");
			save.Indentation++;
			save.WriteElementString("NetworkName", name);
			save.WriteElementString("u0", u0.ToString());
			save.WriteElementString("du0", du0.ToString());
			save.WriteElementString("dt", dt.ToString());
			save.Indentation--;
			save.WriteEndElement();
			// write people data
			save.WriteStartElement("People");
			save.Indentation++;
			foreach (object personobject in people) {
				Person person = (Person) personobject;
				save.WriteStartElement("Person");
				save.Indentation++;
				save.WriteElementString("Name", person.name);
				save.Indentation--;
				save.WriteEndElement();
			}
			save.Indentation--;
			save.WriteEndElement();
			// write meetings data
			save.WriteStartElement("Meetings");
			save.Indentation++;
			foreach (object meetingobject in meetings) {
				Meeting meeting = (Meeting) meetingobject;
				save.WriteStartElement("Meeting");
				save.Indentation++;
				save.WriteElementString("Name", meeting.name);
				save.Indentation--;
				save.WriteEndElement();
			}
			save.Indentation--;
			save.WriteEndElement();
			// write rooms data
			save.WriteStartElement("Rooms");
			save.Indentation++;
			foreach (object roomobject in rooms) {
				Room room = (Room) roomobject;
				save.WriteStartElement("Room");
				save.Indentation++;
				save.WriteElementString("Name", room.name);
				save.WriteElementString("Capacity", room.capacity.ToString());
				save.Indentation--;
				save.WriteEndElement();
			}
			save.Indentation--;
			save.WriteEndElement();
			// write timeslots data
			save.WriteStartElement("TimeSlots");
			save.Indentation++;
			foreach (object timeslotobject in timeslots) {
				TimeSlot timeslot = (TimeSlot) timeslotobject;
				save.WriteStartElement("TimeSlot");
				save.Indentation++;
				save.WriteElementString("Name", timeslot.name);
				save.Indentation--;
				save.WriteEndElement();
			}
			save.Indentation--;
			save.WriteEndElement();

			// write rsvp data
			save.WriteStartElement("RSVP");
			save.Indentation++;
			for (int m = 0; m < meetings.Count; m++) {
				for (int p = 0; p < people.Count; p++) {
					save.WriteElementString("RSVPElement", rsvp[m,p].ToString());
				}
			}
			save.Indentation--;
			save.WriteEndElement();

			// write network data
			save.WriteStartElement("Network");
			save.Indentation++;
			for (int t = 0; t < timeslots.Count; t++) {
				for (int m = 0; m < meetings.Count; m++) {
					for (int r = 0; r < rooms.Count; r++) {
						save.WriteStartElement("NetworkElement");
						save.Indentation++;
						save.WriteElementString("VCurrent", V[t, r, m].ToString());
						save.WriteElementString("VPrevious", VPrevious[t, r, m].ToString());
						save.WriteElementString("uCurrent", u[t, r, m].ToString());
						save.WriteElementString("uInitial", uInitial[t, r, m].ToString());
						save.Indentation--;
						save.WriteEndElement();
					}
				}
			}
			save.Indentation--;
			save.WriteEndElement();

			// close XML file
			save.Indentation--;
			save.WriteEndElement();
			save.WriteEndDocument();
			save.Close();
		}
		#endregion
	}
}
